#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pandas as pd
import re

# The input file is Table. 18-1 from ref2440 (Yaws, 1999) as a scanned
# and OCR-processed document (some species names have been corrected or
# adjusted):
INPUT_TABLE = '~/literat/ref2440-henry-ocr-table-18-1.txt'

def grep_i(searchstring, filename):
    # suffix _i indicates grep option '-i' for case insensitive
    results = []
    with open(filename) as f:
        for line in f:
            if (searchstring.lower() in line.lower()):
                results.append(line.strip())
        return results

def get_casrn_and_inchikey(name):
    # Try to find the current species "name" in tmp_species_calch.txt
    # and extract casrn and inchikey:
    result = grep_i('"'+name+'"', '../tmp_species_calch.txt')
    if (result):
        result_casrn    = re.search('"([0-9]+-[0-9]{2}-[0-9])"', result[0])
        result_inchikey = re.search('"([A-Z]{14}-[A-Z]{10}-[A-Z])"', result[0])
        if (result_casrn and result_inchikey):
            return (result_casrn.group(1), result_inchikey.group(1))
        else:
            return (None, None)
    else:
        return (None, None)

if __name__ == '__main__':

    # https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
    df = pd.read_csv(INPUT_TABLE, skipinitialspace=True)

    f_dat  = open('../input/ref2440.dat', 'w')

    for index, row in df.iterrows():
        n       = int(row['NO'])
        formula = row['FORMULA']
        name    = row['NAME']
        temp    = row['T']
        Hvpx    = row['Hvpx']
        Hvpc    = row['Hvpc']
        # skip all lines that don't contain Henry's law constants:
        if (Hvpx=='---'): continue
        Hvpx = float(Hvpx)
        Hvpc = float(Hvpc)

        if (name=='NONYLPHENOL'): continue # isomer mix

        # check consistency between Hvpx and Hvpc:
        ratio = Hvpx/(55555*Hvpc)
        if (ratio<0.9999 or ratio>1.0001): 
            print(f'ERROR: {n:4}   {Hvpx:10.4E}   {Hvpc:10.4E}   {ratio}  {name}')

        # find casrn and inchikey for name:
        casrn, inchikey = get_casrn_and_inchikey(name)
        if (casrn and inchikey):
            casrn_str = '"' + casrn + '"'
            print(f'{n:4}, {casrn_str:>14}, "{inchikey}", {Hvpx:10.4E}, ' +
                  f'{Hvpc:10.4E}, {temp:>5}, "{name}"', file=f_dat)
        else:
            print(f'ERROR: No casrn and/or inchikey for {n:4} {name}')

    f_dat.close()
