#!/usr/bin/env python3
# -*- coding: utf-8 -*- Time-stamp: <2023-05-23 08:57:17 sander>

import numpy as np
from henry_util import MO, MCl, MH2O, atm, Rgas, CtoK, cH2O, rhoH2O

MCl2  = MCl*2    # kg/mol
MCl2O = MCl*2+MO # kg/mol

def check_Cl2():
    print('Cl2:')
    # temperature range:
    T = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110]) + CtoK

    # data on p. 333 of ref3857:
    x1_ref3857 = np.array([2.48, 1.88, 1.50, 1.23, 1.06, 0.939, 0.849, 
                           0.784, 0.737, 0.697, 0.668])

    # How were these values calculated from Tab. 1 of Adams & Edmonds (ref3925)?
    # Using the trial and error method, it seems that the data were
    # calculated like this:

    # original solubility data from Tab. I of ref3925 at 750 and 800 mmHg:
    sol_750 = np.array([9.65, 7.29, 5.80, 4.77, 4.07, 3.59, 3.23, 2.96, 2.76, 2.60, 2.47])
    sol_800 = np.array([10.21, 7.69, 6.08, 4.99, 4.27, 3.75, 3.37, 3.08, 2.87, 2.69, 2.56])

    # interpolate to 760 Torr:
    sol_760 = 0.8 * sol_750 + 0.2 * sol_800

    # it seems that x1 was calculated with this formula:
    x1_760 = 1E6*sol_760/(MCl2*cH2O*rhoH2O(T))

    # relative difference:
    diff = x1_760/x1_ref3857 - 1

    for i in range(len(T)):
        print(f'{T[i]}: {x1_ref3857[i]:<7} {x1_760[i]:.3f} {diff[i]:>12.4E}')

    # Conclusions:
    #
    # 1) The data refer to an effective Henry solubility that takes into
    #    account the hydrolysis in the aqueous-phase:
    #    Hs_eff = ([Cl2]+[HOCl]) / p(Cl2)
    #
    # 2) Since data at 750 and 800 mmHg were used, the results are valid at
    #    a partial pressure p(Cl2) = 760 mmHg, and not a infinite dilution.

def check_Cl2O():
    print('Cl2O:')
    # The values of x1, L, and S are shown on p. 452. Here is the
    # interconversion between them:
    T  = np.array([ 273.15, 283.15, 293.15 ])
    x1 = np.array([ 0.5259, 0.4273, 0.3353 ])
    L  = np.array([ 1375, 962, 672 ])
    S  = np.array([ 533.3, 359.9, 243.4 ]) / 100
    print('T [K]:               ', T)
    print('S [g(Cl2O)/g(H2O)]:  ', S)
    print('L [dimless]:         ',L)
    print('L calculated from S: ', 1000 * S * Rgas * T / (MCl2O * atm))
    print('x1 [(mol/mol)/atm]:  ', x1)
    x2 = S * MH2O / MCl2O # from g/g to mol/mol
    print('x1 calculated from S:', x2/(x2+1)) # from "per solvent" to "per solution"

if __name__ == '__main__':
    check_Cl2()
    check_Cl2O()
